/**
 * Copyright (c) Members of the EGEE Collaboration. 2007-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *
 */


/*!
    \file   scas_log.c
    \brief  Logging routines for SCAS
    \author Martijn Steenbakkers for the EU DataGrid.
*/

#define _XOPEN_SOURCE	500

/*****************************************************************************
                            Include header files
******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <syslog.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>


#include "scas_log.h"


/******************************************************************************
                       Define module specific variables
******************************************************************************/
static FILE *  scas_logfp=NULL; /*!< stream associated with logfile \internal */
static int     logging_usrlog=0; /*!< flag to do user logging \internal */
static int     logging_syslog=1; /*!< flag to use syslog \internal */
static int     log_level=LOG_NOTICE; /*!< default log level \internal */
static int     should_close_scas_logfp = 0; /*!< Flag to check if the log stream should be closed \internal */ 
static char *  scas_log_path = NULL;  /*!< String that holds the path to the logfile on the file system \internal */

/******************************************************************************
Function:       scas_log_open()
Description:    Start logging
Parameters:
                path:    path of logfile
                logtype: DO_USRLOG, DO_SYSLOG
		facility: LOG_DAEMON etc.
		level:	 LOG_ERR..LOG_DEBUG
Returns:        0 succes
                1 failure
******************************************************************************/
/*!
    \fn scas_log_open(
        unsigned short logtype
	int facility
	int level
        )
    \brief Start logging

    This function should only be used by the SCAS itself.
    \param path    path of logfile.
    \param logtype DO_USRLOG, DO_SYSLOG
    \param facility syslog facility
    \param loglevel LOG_ERR..LOG_DEBUG

    \retval 0 succes.
    \retval 1 failure.
    \internal
*/
int
scas_log_open(char * path, unsigned short logtype, int facility, int level)
{
    char *tmp;

    /* We specify a logtype: reset current setting */
    if (logtype)    {  /* We specify a logtype: reset current setting */
	logging_usrlog=logging_syslog=0;
	if (facility<0)
	    facility=DEFAULT_LOG_FACILITY;
    }

    /* logfile part */
    if ((logtype & DO_USRLOG) == DO_USRLOG)
    {
        if (path != NULL)
        {
	    logging_usrlog=1;
            /* Try to append to file */
	    if (scas_logfp) {
		/* reopening file */
		if ((scas_logfp = freopen(path, "a",scas_logfp)) == NULL)   {
		    scas_log(LOG_ERR, "Reopening log file failed: %s\n",
			    strerror(errno));
		    if (logging_syslog)
			syslog(LOG_ERR,
				"scas_log_open(): Cannot open logfile %s\n",
				path);
		    return 1;
		}
	    } else if ((scas_logfp = fopen(path, "a")) == NULL) {
                fprintf(stderr, "SCAS: scas_log_open(): Cannot open logfile %s: %s\n",
                        path, strerror(errno));
                if (logging_syslog)
                    syslog(LOG_ERR,
			    "scas_log_open(): Cannot open logfile %s\n",
			    path);
                return 1;
            }
            should_close_scas_logfp = 1;
            scas_log_path = path;
        }
        else
        {
	    /* invalid path specified, remove from logtype, only if no open file
	     * exists */
	    if (!scas_logfp)	{
		logging_usrlog=0;
		logtype=DO_SYSLOG;
	    }
        }
    }
    
    /* syslog part */
    if (facility>=0 && (logtype & DO_SYSLOG)==DO_SYSLOG)	{
	logging_syslog=1;
	openlog("scas", LOG_CONS|LOG_PID, facility);
    }

    /* Set loglevel */
    if (level>=0)   {
	log_level=level;
	if (log_level<LOG_ERR)
	    log_level=LOG_ERR;
	else if (log_level>LOG_DEBUG)
	    log_level=LOG_DEBUG;
    
	switch(log_level)   {
	    case LOG_ERR:	tmp="LOG_ERR"; break;
	    case LOG_WARNING:	tmp="LOG_WARNING"; break;
	    case LOG_NOTICE:	tmp="LOG_NOTICE"; break;
	    case LOG_INFO:	tmp="LOG_INFO"; break;
	    case LOG_DEBUG:	tmp="LOG_DEBUG"; break;
	    default:		tmp="unknown"; break;
	}
	scas_log(LOG_DEBUG,"scas_log_open(): loglevel set to %s\n",tmp);
    }

    /* Don't look at the external string value to prefix the log messages. */
    return 0;
}

/******************************************************************************
Function:       scas_log()
Description:    Log information to file and or syslog
Parameters:
                prty:    syslog priority (if 0 don't syslog)
                fmt:     string format
Returns:        0 succes
                1 failure
******************************************************************************/
/*!
    \fn scas_log(
        int prty,
        char * fmt,
        ...
        )
    \brief log information
    
    This function logs information for SCAS and its plugins.
    Syslog() is called with the specified priority. No syslog() is done if the
    priority is 0.

    \param prty syslog priority (if 0 don't syslog).
    \param fmt  string format
    \param ...  variable argument list

    \retval 0 succes.
    \retval 1 failure.
*/
int
scas_log(int prty, char * fmt, ...)
{
    va_list pvar;
    char    buf[MAX_LOG_BUFFER_SIZE];
    int     res;

    time_t      _time;
    struct tm * _time_s;
    char        extra_logstr[MAX_TIME_STRING_SIZE];
    size_t      extra_size;

    if (prty<LOG_ERR)
	prty=LOG_ERR;
    else if (prty>LOG_DEBUG)
	prty=LOG_DEBUG;

    /* Do we need to log this? */
    if (prty>log_level)
	return 0;

    va_start(pvar, fmt);
    res=vsnprintf(buf,(size_t)MAX_LOG_BUFFER_SIZE,fmt,pvar);
    va_end(pvar);
    if ( (res >= MAX_LOG_BUFFER_SIZE) || (res < 0) )
    {
        fprintf(stderr,"SCAS: scas_log(): log string too long (> %d)\n",
                MAX_LOG_BUFFER_SIZE);
    }
    if (logging_usrlog)
    {
        if (scas_logfp == NULL)
        {
            fprintf(stderr,"SCAS: scas_log() error: cannot open file descriptor\n");
            return 1;
        }

	_time   = time(NULL);
	extra_size = 0;
    
	_time_s    = localtime(&_time);
	if (_time_s==NULL)
	    extra_logstr[0]='\0';
	else {
	    /* Note: strftime() returns number of bytes WITHOUT \0 */
	    extra_size = strftime(extra_logstr, (size_t)MAX_TIME_STRING_SIZE,
				 "%Y-%m-%d.%H:%M:%S%z", _time_s);
	    if (extra_size==0)
		extra_logstr[0]='\0';
	}

        if (extra_size == 0)
            fprintf(scas_logfp,"SCAS %d[%ld]: %s",
		    prty-LOG_ERR+1, (long)getpid(), buf);
        else
            fprintf(scas_logfp,"SCAS %d[%ld]: %s : %s",
		    prty-LOG_ERR+1, (long)getpid(), extra_logstr, buf);
        fflush(scas_logfp);
    }
    if (logging_syslog)
    {
        syslog(prty, "%s", buf);
    }

    return 0;
}


/******************************************************************************
Function:       scas_log_close()
Description:    Stop logging
Parameters:
Returns:        0 succes
                1 failure
******************************************************************************/
/*!
    \fn scas_log_close()
    \brief Stop logging.

    \internal
    This function should only be used by the SCAS itself.

    \retval 0 succes.
    \retval 1 failure.
    \internal
*/
int
scas_log_close()
{
    if (should_close_scas_logfp)
    {
        fclose(scas_logfp);
        scas_logfp=NULL;
    }

    return 0;
}



/******************************************************************************
Function:       scas_get_log_file_pointer()
Description:    Get the file pointer to the logfile
Parameters:
Returns:        FILE * pointer to the open file
******************************************************************************/
FILE * scas_get_log_file_pointer(void)
{
    return scas_logfp;
}


/******************************************************************************
Function:       scas_get_log_type()
Description:    Get the active type of logging facility (syslog and/or file logging)
Parameters:
Returns:        unsigned short
******************************************************************************/
unsigned short scas_get_log_type (void)
{
    if (logging_syslog && logging_usrlog) return DO_SYSLOG|DO_USRLOG;
    if (logging_usrlog) return DO_USRLOG;
    if (logging_syslog) return DO_SYSLOG;

    return 0;
}

/******************************************************************************
Function:       scas_get_log_path()
Description:    Get the path string to the logfile
Parameters:
Returns:        char *
******************************************************************************/
char * scas_get_log_path (void)
{
    return scas_log_path;
}


/******************************************************************************
Function:       scas_get_log_path()
Description:    Converts strings representation of log facility to int
Parameters:	char *
Returns:        int
******************************************************************************/
int scas_get_log_facility(const char *log_facility)
{
    if (log_facility==NULL)
	return DEFAULT_LOG_FACILITY;

    if (strcmp(log_facility,"LOG_AUTH")==0)	return LOG_AUTH;
    if (strcmp(log_facility,"LOG_CRON")==0)	return LOG_CRON;
    if (strcmp(log_facility,"LOG_DAEMON")==0)	return LOG_DAEMON;
    if (strcmp(log_facility,"LOG_KERN")==0)	return LOG_KERN;
    if (strcmp(log_facility,"LOG_LOCAL0")==0)	return LOG_LOCAL0;
    if (strcmp(log_facility,"LOG_LOCAL1")==0)	return LOG_LOCAL1;
    if (strcmp(log_facility,"LOG_LOCAL2")==0)	return LOG_LOCAL2;
    if (strcmp(log_facility,"LOG_LOCAL3")==0)	return LOG_LOCAL3;
    if (strcmp(log_facility,"LOG_LOCAL4")==0)	return LOG_LOCAL4;
    if (strcmp(log_facility,"LOG_LOCAL5")==0)	return LOG_LOCAL5;
    if (strcmp(log_facility,"LOG_LOCAL6")==0)	return LOG_LOCAL6;
    if (strcmp(log_facility,"LOG_LOCAL7")==0)	return LOG_LOCAL7;
    if (strcmp(log_facility,"LOG_LPR")==0)	return LOG_LPR;
    if (strcmp(log_facility,"LOG_MAIL")==0)	return LOG_MAIL;
    if (strcmp(log_facility,"LOG_NEWS")==0)	return LOG_NEWS;
    if (strcmp(log_facility,"LOG_USER")==0)	return LOG_USER;
    if (strcmp(log_facility,"LOG_UUCP")==0)	return LOG_UUCP;
   
#ifdef LOG_AUDIT
    if (strcmp(log_facility,"LOG_AUDIT")==0)	return LOG_AUDIT;
#endif
#ifdef LOG_AUTHPRIV
    if (strcmp(log_facility,"LOG_AUTHPRIV")==0)	return LOG_AUTHPRIV;
#endif
#ifdef LOG_FTP
    if (strcmp(log_facility,"LOG_FTP")==0)	return LOG_FTP;
#endif
#ifdef LOG_SYSLOG
    if (strcmp(log_facility,"LOG_SYSLOG")==0)	return LOG_SYSLOG;
#endif

    return DEFAULT_LOG_FACILITY;
}


/******************************************************************************
CVS Information:
    $Source: /srv/home/dennisvd/svn/mw-security/scas/src/scas-server/logging/scas_log.c,v $
    $Date: 2010-04-29 10:59:40 $
    $Revision: 1.10 $
    $Author: okoeroo $
******************************************************************************/
