/**
 * Copyright (c) Members of the EGEE Collaboration. 2007-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *
 */

#define _XOPEN_SOURCE	500

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <limits.h>
#include <errno.h>

#include "scas_config.h"

#include "daemonize.h"

#include "common.h"
#include "pdp_xacml_authz_process.h"

#include "xacml_server.h"
#include "xacml_io_ssl.h"
#include "ssl-common.h"
#include "net_common.h"


#include "open_conf_file.h"


#include "scas_log.h"


#if 0
#define FORKCHILDMAX 10
#endif

/* After child has ended due to signal more than MAX_CRASHES times
 * consecutively in MAX_CRASHTIME, we stop the service */
#define MAX_CRASHES	3

#define MAX_CRASHTIME	60

#if 0
#define ENABLE  1
#define DISABLE 0
#define TRUE    1
#define FALSE   0
#endif


#define SCAS_ERR_CMDLINE     10
#define SCAS_ERR_LOG_FAIL    20
#define SCAS_ERR_CONFIG_FAIL 30
#define SCAS_ERR_UNPRIV_PORT 40
#define SCAS_ERR_SYSCALLS    50
#define SCAS_ERR_PIDFILE     60

#define SCAS_PROCESS_UNINIT  0
#define SCAS_PROCESS_PARENT  1001
#define SCAS_PROCESS_CHILD   1002

#define SCAS_SERVICE_REFRESH_TIMER 600  /* In seconds */

#ifndef SCAS_PIDFILE
# define SCAS_PIDFILE	"/var/run/scas/scas.pid"
#endif


/* defined extern in interface/pdp_xacml_authz_process.h */
pthread_mutex_t mutex;

static int parentprocess = SCAS_PROCESS_UNINIT;
static int mychild       = 0;

static int done = 0;
static int gsig = 0;
static int reinitlog = 0;   /* whether we need to reinit the logfile */
static int waitchild = 0;   /* whether we have a child that finished */
static int killchild = 0;   /* whether we need to kill the child */


static long service_refresh_timer = 0;


/* prototype */

/*int process_xacml (int listenport, int secured);*/
static int process_xacml (int listenport);
void sigintaction(int sig, siginfo_t *act, void *context);
static void setup_signal_handlers(void);
static int killwaitchild(int childpid,int *crashcounter,time_t *crashstart);
static int checkpoint(void);
static int reinitlogfile(void);
static void usage (void);
static int write_pidfile(void);


int reinitlogfile()
{
    unsigned short scas_logfile_type = 0;
    char * scas_logfile_path = NULL;

    /* Clear flag */
    reinitlog=0;

    /* Get file pointer of the logfile and store the file descriptor */ 
    /* Only do this action if logging to file, skip procedure when logging to SYSLOG */
    scas_logfile_type = scas_get_log_type();
    if ((scas_logfile_type & DO_USRLOG) !=0)
    {
        scas_logfile_path = scas_get_log_path();
   
        scas_log(LOG_DEBUG, "SCAS: Re-opening the log file %s\n", scas_logfile_path);
    
        /* Re-open log file */
        if (scas_log_open (scas_logfile_path, scas_logfile_type, -1,-1))
        {
            fprintf (stderr, "SCAS logging facility failed to start up! Tried using the log file: %s\n",
		    scas_logfile_path);
            return SCAS_ERR_LOG_FAIL;
        }
    }

    return 0;
}

/*
 * Signal handler
 * This signal handler will stop the service gracefully
 */
void sigintaction(int sig, siginfo_t *info, void *context)
{
    switch (sig)
    {
        case SIGTERM :
            gsig = sig;
            /* Stop the service */
            if (SCAS_PROCESS_PARENT == parentprocess)	{
		/* parent */
		killchild=1;
		done = 1;
	    } else {
		/* child: time is up... */
		_exit(128+SIGTERM);
	    }
            break;
        case SIGINT :
            gsig = sig;
	    done = 1;
            /* Stop the service */
            if (SCAS_PROCESS_PARENT == parentprocess)	{
		killchild=1;
	    }
            break;
        case SIGHUP :
            /* Only the parent needs to reinit the log and re-issue a child process */
            if (SCAS_PROCESS_PARENT == parentprocess)	{
		gsig = sig;
		killchild=1;
		reinitlog=1;
	    }
            break;
        case SIGCHLD :
            /* Need to waitchild */
            if (SCAS_PROCESS_PARENT == parentprocess)	{
		gsig = sig;
		if (info->si_pid==mychild)
		    waitchild=1;
	    }
            break;
        default :
	    break;
    }
}

static void setup_signal_handlers(void)	{
    /* Used to set signal handlers */
    struct sigaction act;

    /* Initialize sigaction struct */
    act.sa_handler=NULL;
    act.sa_sigaction=&sigintaction;

    /* Block all signals that we catch */
    sigemptyset(&act.sa_mask);
    sigaddset(&act.sa_mask,SIGTERM);
    sigaddset(&act.sa_mask,SIGINT);
    sigaddset(&act.sa_mask,SIGCHLD);
    sigaddset(&act.sa_mask,SIGHUP);

    /* Do not react on a stopped child */
    act.sa_flags=SA_SIGINFO | SA_NOCLDSTOP;

    /* Set handler for the different signals */
    sigaction(SIGTERM,&act,NULL);
    sigaction(SIGINT ,&act,NULL);
    sigaction(SIGCHLD,&act,NULL);
    sigaction(SIGHUP ,&act,NULL);
}

/**
 * waits for child and optionally kills it (when global killchild has been set
 * by signal handler).
 */
static int killwaitchild(int childpid,int *crashcounter,time_t *crashstart) {
    const struct timespec mysleep={0,100000000L};
    const int inttimeout=4;
    const int totaltimeout=6;
    int rc;
    int status = 0;
    int step;

    /* reset waitchild flag */
    waitchild=0;

    /* Get status for child */
    rc=waitpid(-1,&status,WNOHANG);

    /* Only kill when there is a child and we need to kill it */
    if (killchild && rc>=0) {
	/* Reset killchild flag */
	killchild=0;
    
	/* We now start a series of kill's: first a SIGINT: normally this should
	 * terminate the child. If it doesn't, then after 20 * 0.1 sec, a
	 * SIGTERM is sent. This should let the child call _exit(). If that
	 * still doesn't work after 20 * 0.1 seconds, it is sent a SIGKILL.
	 * We will wait with a WNOHANG for the child, except after the
	 * SIGKILL. */
	/* Kill child with SIGINT */
	scas_log(LOG_DEBUG, "Sending SIGINT to child %d.\n",childpid);
	rc = kill(childpid, SIGINT);
	if (rc<0)	{
	    /* kill failed: child might have died just before */
	    if (errno!=ESRCH)	{
		scas_log(LOG_ERR, "Cannot send SIGINT to child %d: %s\n",
			childpid,strerror(errno));
		return 1;
	    }
	    /* Child died just before kill, check wait status here */
	    scas_log(LOG_INFO, "Child process no longer exists\n");
	    rc=waitpid(-1,&status,WNOHANG);
	} else	{
	    /* sent SIGINT succesfully, wait status. Loop for totaltimeout
	     * steps */
	    for (step=0; step<=totaltimeout*10; step++)   {
		rc=waitpid(-1,&status,WNOHANG);
		/* End loop when child has finished or when we cannot waitpid */
		if (rc>0 || (rc<0 && errno!=EINTR))
		    break;
		/* After 2 seconsd, send a SIGTERM */
		if (step==inttimeout*10)	{
		    /* Kill child with SIGTERM */
		    scas_log(LOG_WARNING,
			"Child hasn't ended after %d sec., sending SIGTERM\n",
			inttimeout);
		    rc = kill(childpid, SIGTERM);
		    if (rc<0)	{
			/* kill failed: child might have died just before */
			if (errno!=ESRCH)   {
			    scas_log(LOG_ERR,
				    "Cannot send SIGTERM to child %d: %s\n",
				    childpid,strerror(errno));
			    return 1;
			}
			/* Child died just before kill, check wait status here */
			scas_log(LOG_NOTICE, "Child process no longer exists\n");
			rc=waitpid(-1,&status,WNOHANG);
			break;
		    }
		} else if (step==totaltimeout*10)    {
		    scas_log(LOG_WARNING,
			"Child hasn't ended after %d sec., sending SIGKILL\n",
			totaltimeout);
		    /* Kill child with SIGTERM */
		    rc = kill(childpid, SIGKILL);
		    if (rc<0)	{ /* kill failed: child might have died just before */
			if (errno!=ESRCH)   {
			    scas_log(LOG_ERR,
				    "Cannot send SIGKILL to child %d: %s\n",
				    childpid,strerror(errno));
			    return 1;
			}
			scas_log(LOG_NOTICE,
				"Child process no longer exists\n");
		    }
		    /* Now child must have terminated */
		    rc=wait(&status);
		    break;
		} else
		    /* Sleep for 0.1 seconds */
		    nanosleep(&mysleep,NULL);
	    }
	}
    }

    /* Check return status of wait(pid) */
    if (rc<0)   {   /* waitpid failed */
	switch (errno)	{
	    case ECHILD:
		scas_log(LOG_NOTICE, "No terminated child is waiting\n");
		break;
	    case EINTR: /* this should not be possible: wait() is in a loop and
			   waitpid has a WNOHANG */
		scas_log(LOG_ERR, "wait got interrupted: %s\n",
			strerror(errno));
		return 1;
	    case EINVAL:
	    default:
		scas_log(LOG_ERR, "Waiting for child failed: %s\n",
			strerror(errno));
		return 1;
	}
    } else if (rc>0)	{   /* waitpid found child */
	if (WIFEXITED(status))	{
	    scas_log(WEXITSTATUS(status) ? LOG_WARNING : LOG_NOTICE,
		    "Child %d ended normally with exit value %d.\n",
		     rc,WEXITSTATUS(status));
	    if (WEXITSTATUS(status)==0)	{
		/* Reset crashcounter */
		*crashcounter=0;
	    } else {
		/* When exitvalue is non-zero: interpret as crash */
		if (*crashcounter==0)
		    if (time(crashstart)==(time_t)-1)
			scas_log(LOG_ERR,"Cannot get time for first crash.\n");
		(*crashcounter)++;
	    }
	} else if (WIFSIGNALED(status))	{
	    scas_log(LOG_WARNING, "Child %d ended due to signal %d.\n",
		     rc,WTERMSIG(status));
	    /* Keep track of number of crashes */
	    if (*crashcounter==0)
		if (time(crashstart)==(time_t)-1)
		    scas_log(LOG_ERR,"Cannot get time for first crash.\n");
	    (*crashcounter)++;
	} else
	    scas_log(LOG_ERR, "Child %d ended for unknown reason.\n",
		     rc);
    } else
	/* waitpid found no child changed status:
	 * we shouldn't have been called */
	scas_log(LOG_NOTICE, "Unexpected: no children needed waiting\n");

    return 0;
}

static int checkpoint()
{
    const long realtime_timeout = service_refresh_timer; /* seconds */
    time_t start,crashstart,curtime;
    int crashcounter=0;

    while (!done)   {
        start = 0;

	scas_log(LOG_DEBUG,"Forking new child process.\n");
	/* Make sure to reset flags just before fork() */
	killchild=waitchild=0;
	mychild  = fork(); /* Use global mychild pid var, to make sure we'll
			      always have a consistent idea about the PID */

        if (mychild == 0)
        {
            /* I'm the child process */
            /* I will (re)start working */
            parentprocess = SCAS_PROCESS_CHILD;

            /* Ignore the SIGHUP call explicitly from a child */
            /* A child process will inherit the registered signal handlers, 
             * which is unwanted behavior */
            signal (SIGHUP, SIG_IGN);
            
	    scas_log(LOG_DEBUG,"Child %ld starting.\n",(long)getpid());

            return 0;
        }
        else if (mychild < 0)
        {
            /* something went wrong... need to bail out */

            scas_log (LOG_ERR, "SCAS: Unable to gracefully restart due to low resources\n");
            exit (1);
        }
        else
        {
            /* I'm the parent */
            /* I'll wait for a signal, like SIGUSR1, to restart a new child as a replacement */
            /* wait (&status); */

            parentprocess = SCAS_PROCESS_PARENT;

            /* Timer */
            start = time(NULL);
	    scas_log(LOG_DEBUG,"Waiting %ld sec before child process refresh\n",
		    realtime_timeout);
	    /* Loop until we either get a signal and killchild or waitchild is
	     * set, or until the timer has expired. In either case, we end the
	     * loop using a break; */
            while ( 1 )
            {	
		/* sleep will either expire or get interrupted */
                if (sleep (10)!=0)
		    scas_log(LOG_DEBUG,"sleep got interrupted\n");

		/* check if either flag got raised */
		if (killchild || waitchild) {
		    /* Need to kill or wait for child */
		    scas_log(LOG_NOTICE,"Caught signal %d...\n",gsig);
		    if (killwaitchild(mychild,&crashcounter,&crashstart)!=0)
			return 1;

		    /* When we got a SIGHUP, also reinit the log */
		    if (reinitlog)
			reinitlogfile(); 

		    break;
		} else if ( (time(NULL) - start) > realtime_timeout)	{
		    /* maximum time, need to refresh the child */
		    scas_log(LOG_DEBUG,"Time to refresh child.\n");
		    killchild=1;
		    if (killwaitchild(mychild,&crashcounter,&crashstart)!=0)
			return 1;
		    break;
		}
            }
	    /* Check we don't have more than MAX_CRASHES successive crashes */
	    if (crashcounter>MAX_CRASHES)   {
		/* Determine interval for the crashes */
		if (time(&curtime)==(time_t)-1)
		    scas_log(LOG_ERR,"Cannot get time\n");
		/* If in shorter than max interval: we need to quit */
		if (curtime-crashstart<MAX_CRASHTIME)   {
		    scas_log(LOG_ERR,
			    "Reached maximum of %d consecutive crashes "
			    "in %d seconds, stopping service.\n",
			    MAX_CRASHES,MAX_CRASHTIME);
		    done=1;
		} else {
		    /* Need to reset the crashes: max time interval has expired.
		     * */
		    scas_log(LOG_WARNING,
			    "Seen more than than %d consecutive crashes, "
			    "but not in less than %d seconds.\n",
			    MAX_CRASHES,MAX_CRASHTIME);
		    /* Reset counter to first crash and corresponding time */
		    crashcounter=1;
		    crashstart=curtime;
		}
	    }
        }
    }

    return 0;
}


static void usage (void)
{
    printf ("\n");
    printf (PACKAGE_STRING);
    printf ("\n\n");
    printf ("   Command line options:\n");
    printf ("   scas [-h|--h] [--conf <configuration file>] [--port <port number>] [--log <path to logfile>] [--cert <public certificate>] [--key <private key file>] [--capath <CA directory>] [--refresh-time <child process refresh interval in seconds>] [--no-daemon]\n");
    printf ("\n");
    printf ("                 -h|--h - The usage output you are looking at right now.\n");
    printf ("     --conf <conf file> - Configuration file for the SCAS service.\n");
    printf ("                          It is advised to primarily use the scas.conf.\n");
    printf ("                          See manpage for more about the the scas.conf options\n");
    printf ("   --port <port number> - TCP/IP port number to listen on.\n");
    printf ("--log <path to logfile> - Sets the path to the log file that will be used.\n");
    printf ("   --cert <public cert> - Sets the public certificate file for the SCAS service.\n");
    printf ("  --key <priv key file> - Sets the private key file associated by the public.\n");
    printf ("                          certificate file.\n");
    printf ("      --capath <CA dir> - Sets the Certificate Authority (CA) directory to be\n");
    printf ("                          used for the CA public certificates and CRL files.\n");
    printf (" --refresh-timer <secs> - Sets the child process refresh interval in seconds.\n");
    printf ("                          The parent process will wait this amount of seconds\n");
    printf ("                          before refreshing the child process for a new one to\n");
    printf ("                          keep memory leakage to a minimum.\n");
    printf ("            --no-daemon - Keeps the service from daemonizing and refreshing.\n");
    printf ("                          This option is very useful when debugging in GDB\n");
    printf ("\n");
    printf ("Note: The command line options override the settings of the configuration file.\n");
    printf ("\n");
}



/* MAIN */
int main(int argc, char *argv[])
{
    int rc = 0;
    int i = 0;
    int portnumber = 0;
    int disableDaemonize = 0;
    int use_cert_chain = -1; /* No default here */
    char * log_filepath  = NULL;
#ifdef SCAS_CONFIG_FILE
    char * conf_filepath = SCAS_CONFIG_FILE;
#else
    char * conf_filepath = NULL;
#endif
    unsigned short scas_log_type = 0;
    int scas_log_facility = -1;
    int scas_log_level = -1;

#ifdef HAVE_XACML_SERVER_SET_TIMEOUTS
    int scas_send_timeout = -1;
    int scas_receive_timeout = -1;
    int scas_accept_timeout = -1;
#endif

    config_name_values cnv = NULL;
    size_t n_config_name_values = 0;

    char *tmpval = NULL;
    long value;


    xacml_io_certFilePub  = NULL;
    xacml_io_certFilePriv = NULL;
    xacml_io_certCAPath   = NULL;



    /* Commandline opts */
    for (i = 1; i < argc; i++)
    {
        if ((strcasecmp (argv[i], "--conf") == 0) && (i + 1 < argc))
        {
            conf_filepath = argv[++i]; 
        }
        else if (strcasecmp (argv[i], "--no-daemon") == 0)
        {
            disableDaemonize = 1;
        }
        else if (strcasecmp (argv[i], "--use-cert-chain") == 0)
        {
            use_cert_chain = 1;
        }
        else if ((strcasecmp (argv[i], "--port") == 0) && (i + 1 < argc))
        {
	    errno=0;
	    tmpval=argv[++i];
	    value=strtol(tmpval,NULL,10);
	    if (errno!=0 || value<1 || value>65535) {
		fprintf(stderr,"SCAS: Value %s is not a valid port number\n",tmpval);
                return SCAS_ERR_CMDLINE;
	    }
	    portnumber=(int)value;
        }
        else if ((strcasecmp (argv[i], "--log") == 0) && (i + 1 < argc))
        {
            log_filepath = argv[++i]; 
        }
        else if ((strcasecmp (argv[i], "--key") == 0) && (i + 1 < argc))
        {
            xacml_io_certFilePriv = argv[++i]; 
        }
        else if ((strcasecmp (argv[i], "--cert") == 0) && (i + 1 < argc))
        {
            xacml_io_certFilePub = argv[++i]; 
        }
        else if ((strcasecmp (argv[i], "--capath") == 0) && (i + 1 < argc))
        {
            xacml_io_certCAPath  = argv[++i]; 
        }
        else if ((strcasecmp (argv[i], "--refresh-timer") == 0) && (i + 1 < argc))
        {
            char * refresh_timer_tmp = argv[++i];

	    errno=0;
            service_refresh_timer = strtol(refresh_timer_tmp, NULL, 10);
	    if (errno!=0)
            {
                fprintf (stderr, "SCAS: Couldn't parse --refresh-timer value. Doesn't seem to be a number: %s\n", refresh_timer_tmp);
                return SCAS_ERR_CMDLINE;
            }
            else if (service_refresh_timer < 1)
            {
                 fprintf (stderr, "SCAS: --refresh-timer value too low: %ld\n", service_refresh_timer);
                 return SCAS_ERR_CMDLINE;
            }
        }
        else if ( (strcasecmp (argv[i], "--h") == 0) || (strcasecmp (argv[i], "-h") == 0) )
        {
            usage();
            return 0;
        }
        else
        {
            /* Unsupported option provided */
            fprintf (stderr, "SCAS: Unsupported commandline option found: %s\n", argv[i]);
            return SCAS_ERR_CMDLINE;
        }
    }

    /* Open and read configuration file, if a configuration file is specified */
    if (conf_filepath)
    {
        if (open_scas_config_file(conf_filepath, &cnv, &n_config_name_values))
        {
            /* scas_log(LOG_ERR, "Couldn't open the configuration file %s\n", conf_filepath); */
            fprintf (stderr, "SCAS: Couldn't open the configuration file %s\n", conf_filepath);
            return SCAS_ERR_CONFIG_FAIL;
        }
    }
    else
    {
        fprintf(stderr, "SCAS: No configuration file specified, only using commandline options\n");
    }


    /* Setting values that are not well expressed in environment variables here. 
       The origin of the value is the configuration file */
    /* If there is no commandline parameter set for the log file destination, then we'll look in the config file */
    if (!log_filepath)
    {
        log_filepath = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_log_file");
	/* Don't do anything when not specified: if not specified to log to
	 * file, use syslog */
/*        {
            fprintf (stderr, "SCAS: Couldn't find the setting for the log file in either the configuration file or the commandline parameters.\n");
            return SCAS_ERR_CONFIG_FAIL;
        }*/
    }



    /* Try to get scas_log_type */
    if (!scas_log_type)
    {
        char * tmp_log_type = NULL;
	
	if ((tmp_log_type=get_scas_config_value_by_name (cnv, n_config_name_values, "scas_log_type")))
	{
	    if (strcmp(tmp_log_type,"syslog")==0)
		scas_log_type=DO_SYSLOG;
	    else if (strcmp(tmp_log_type,"file")==0)
		scas_log_type=DO_USRLOG;
	    else if (strcmp(tmp_log_type,"both")==0)
		scas_log_type=DO_SYSLOG | DO_USRLOG;
	    else    {
		fprintf(stderr,"SCAS: Value for scas_log_type not understood: %s\n",tmp_log_type);
		return SCAS_ERR_CONFIG_FAIL;
	    }
	}
    }
    
    /* Check if scas_log_type is specified: if not depends on logfile */
    if (scas_log_type==0)
	scas_log_type=(log_filepath ? DO_USRLOG : DO_SYSLOG);

    /* Check combination of log_filepath and scas_log_type */
    if (log_filepath==NULL) {
	if (scas_log_type==DO_USRLOG)	{
	    fprintf(stderr,"SCAS: Value of scas_log_type is \"file\" but \"scas_log_file\" is not set\n");
	    return SCAS_ERR_CONFIG_FAIL;
	}
	if (scas_log_type & DO_USRLOG)	{
	    fprintf(stderr,"SCAS: Value of scas_log_type includes logging to file, "
			"but \"scas_log_file\" is not set, using only syslog\n");
	    scas_log_type=DO_SYSLOG;
	}
    }

    /* Get log facility if we include logging to syslog */
    if (scas_log_type & DO_SYSLOG)  {
	char *tmp_log_facility=NULL;
	
	if ((tmp_log_facility=get_scas_config_value_by_name (cnv, n_config_name_values, "scas_log_facility")))
	    /* Convert string to facility, or returns default */
	    scas_log_facility=scas_get_log_facility(tmp_log_facility);
    }

    /* Get loglevel */
    if ((tmpval=get_scas_config_value_by_name (cnv, n_config_name_values, "scas_debug_level")))
    {
	errno=0;
	value=strtol(tmpval,NULL,10);
	if (errno!=0 || value<0 || value>5) {
	    fprintf(stderr,"SCAS: Cannot convert %s to a value [0:5]\n",tmpval);
	    return SCAS_ERR_CONFIG_FAIL;
	}
	scas_log_level=LOG_CRIT+(int)value;
    }

    /* Open Log file facility */
    if (scas_log_open (log_filepath, scas_log_type, scas_log_facility, scas_log_level))
    {
        fprintf (stderr, "SCAS logging facility failed to start up! Tried using the log file: %s\n", log_filepath);
        return SCAS_ERR_LOG_FAIL;
    }

    /* Test routines by printing the structures */
/*    print_config_name_value (cnv, n_config_name_values);*/


    /* setup of LCAS and LCMAPS environment variables */
    if (setup_scas_config_name_value (cnv, n_config_name_values))
    {
        fprintf (stderr, "SCAS: Couldn't setup LCAS and LCMAPS configuration file options into the environment of this process.\n");
        scas_log(LOG_ERR,           "Couldn't setup LCAS and LCMAPS configuration file options into the environment of this process.\n");
        return SCAS_ERR_CONFIG_FAIL;
    }

    /* SCAS service child process refresh rate selection */
    if (service_refresh_timer == 0)
    {
        char * tmp_refresh_rate = NULL;

        if (!(tmp_refresh_rate = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_refresh_rate")))
        {
            service_refresh_timer = SCAS_SERVICE_REFRESH_TIMER;
        }
        else
        {
	    errno=0;
            service_refresh_timer = strtol(tmp_refresh_rate, NULL, 10);
            if (errno!=0)
            {
                fprintf (stderr, "SCAS: Couldn't parse scas_refresh_rate value in the configuration file. Doesn't seem to be a number: %s\n", tmp_refresh_rate);
                scas_log (LOG_ERR, "SCAS: Couldn't parse scas_refresh_rate value in the configuration file. Doesn't seem to be a number: %s\n", tmp_refresh_rate);
                return SCAS_ERR_CONFIG_FAIL;
            }
        }
        scas_log(LOG_DEBUG, "Using a child process refresh interval of %ld seconds\n", service_refresh_timer);
    }

    /* Send, receive and accept timeout values for gsoap */
    /* send */
    if ((tmpval = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_send_timeout")))
    {
#ifdef HAVE_XACML_SERVER_SET_TIMEOUTS
	errno=0;
	value=strtol(tmpval,NULL,10);
	if (errno!=0 || value<0 || value>INT_MAX) {
	    fprintf(stderr,
		    "SCAS: Cannot convert value %s for scas_send_timeout to a value [0:INT_MAX]\n",
		    tmpval);
	    scas_log(LOG_ERR,
		    "SCAS: Cannot convert value %s for scas_send_timeout to a value [0:INT_MAX]\n",
		    tmpval);
	    return SCAS_ERR_CONFIG_FAIL;
	}
	scas_send_timeout=(int)value;
#else
	scas_log(LOG_ERR,"Found scas_send_timeout, but SCAS is built without support for it.\n");
#endif
    }
   
    /* receive */
    if ((tmpval = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_receive_timeout")))
    {
#ifdef HAVE_XACML_SERVER_SET_TIMEOUTS
	errno=0;
	value=strtol(tmpval,NULL,10);
	if (errno!=0 || value<0 || value>INT_MAX) {
	    fprintf(stderr,
		    "SCAS: Cannot convert value %s for scas_receive_timeout to a value [0:INT_MAX]\n",
		    tmpval);
	    scas_log(LOG_ERR,
		    "SCAS: Cannot convert value %s for scas_receive_timeout to a value [0:INT_MAX]\n",
		    tmpval);
	    return SCAS_ERR_CONFIG_FAIL;
	}
	scas_receive_timeout=(int)value;
#else
	scas_log(LOG_ERR,"Found scas_receive_timeout, but SCAS is built without support for it.\n");
#endif
    }
   
    /* accept */
    if ((tmpval = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_accept_timeout")))
    {
#ifdef HAVE_XACML_SERVER_SET_TIMEOUTS
	errno=0;
	value=strtol(tmpval,NULL,10);
	if (errno!=0 || value<0 || value>INT_MAX) {
	    fprintf(stderr,
		    "SCAS: Cannot convert value %s for scas_accept_timeout to a value [0:INT_MAX]\n",
		    tmpval);
	    scas_log(LOG_ERR,
		    "SCAS: Cannot convert value %s for scas_accept_timeout to a value [0:INT_MAX]\n",
		    tmpval);
	    return SCAS_ERR_CONFIG_FAIL;
	}
	scas_accept_timeout=(int)value;
#else
	scas_log(LOG_ERR,"Found scas_accept_timeout, but SCAS is built without support for it.\n");
#endif
    }

#ifdef HAVE_XACML_SERVER_SET_TIMEOUTS
    /* Set the gsoap timeouts (or keep their defaults) */
    xacml_server_set_timeouts(scas_send_timeout,scas_receive_timeout,scas_accept_timeout);
#endif

    /* Whether to use cert-chain subject attribute */
    if (use_cert_chain==-1)    { /* when not yet set, check config */
	char * tmp_choice = NULL;

	if ((tmp_choice = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_use_cert_chain")))
	{
	    if (strcasecmp ("yes", tmp_choice) == 0 ||
	        strcasecmp ("true", tmp_choice) == 0)
	    {
		use_cert_chain = 1;
	    } else if (strcasecmp ("no", tmp_choice) == 0 ||
		       strcasecmp ("false", tmp_choice) == 0)
	    {
		use_cert_chain = 0;
	    } else {
		scas_log(LOG_ERR,
			"SCAS: Invalid value \"%s\" for scas_use_cert_chain\n",
			tmp_choice);
		return SCAS_ERR_CONFIG_FAIL;
	    }
	}
    }
    /* Set the value (or keep default) and log it. */
    switch (set_use_cert_chain(use_cert_chain))	{
	case 0:
	    scas_log(LOG_DEBUG, "Using DN+FQANs for LCMAPS\n");
	    break;
	case 1:
	    scas_log(LOG_DEBUG, "Using cert-chain for LCMAPS (when available)\n");
	    break;
	default:
	    scas_log(LOG_WARNING, "Unknown value %d for use_cert_chain\n",
		    set_use_cert_chain(-1));
	    break;
    }

    /* Daemonization */
    /* Disable the daemonization (for debugging purposes) */
    if (!disableDaemonize)
    {
	char * tmp_choice = NULL;

	if ((tmp_choice = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_no_daemonize")))
	{
	    if (strcasecmp ("yes", tmp_choice) == 0 ||
	        strcasecmp ("true", tmp_choice) == 0)
	    {
		disableDaemonize = 1;
		scas_log(LOG_DEBUG, "Using \"no daemonizing\" mode\n");
	    }
	}
    } 

    /* Portnumber selection */
    if (portnumber == 0)
    {
        char * tmp_port_num = NULL;

        if (!(tmp_port_num = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_port")))
        {
            fprintf (stderr, "SCAS: Couldn't find the setting for the listening portnumber in the configuration file or the commandline parameters.\n");
            scas_log (LOG_ERR, "SCAS: Couldn't find the setting for the listening portnumber in the configuration file or the commandline parameters.\n");
            return SCAS_ERR_CONFIG_FAIL;
        }
        else
        {
	    errno=0;
	    value=strtol(tmp_port_num,NULL,10);
	    if (errno!=0 || value<1 || value>65535) {
		fprintf(stderr,"SCAS: Value %s is not a valid port number\n",tmp_port_num);
		scas_log(LOG_ERR,"Value %s is not a valid port number\n",tmp_port_num);
                return SCAS_ERR_CMDLINE;
	    }
	    portnumber=(int)value;
            scas_log(LOG_DEBUG, "Using portnumber %d\n", portnumber);
        }
/*        free (tmp_port_num);*/ /* Will be freed by free_config_name_value
 */
    }

    /* CA path */
    if (!xacml_io_certCAPath)
    {
        if (!(xacml_io_certCAPath = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_capath")))
        {
            fprintf (stderr, "SCAS: Couldn't find the setting for the CA Path which holds the CA certificates in the configuration file or the commandline parameters.\n");
            scas_log (LOG_ERR, "SCAS: Couldn't find the setting for the CA Path which holds the CA certificates in the configuration file or the commandline parameters.\n");
            return SCAS_ERR_CONFIG_FAIL;
        }
        else
            scas_log(LOG_DEBUG, "Using CA Path: %s\n", xacml_io_certCAPath); 
    } 

    /* Host / service certificate */
    if (!xacml_io_certFilePub)
    {
        if (!(xacml_io_certFilePub = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_hostcert")))
        {
            fprintf (stderr, "SCAS: Couldn't find the setting for the Host / Service certificate file in the configuration file or the commandline parameters.\n");
            scas_log (LOG_ERR, "SCAS: Couldn't find the setting for the Host / Service certificate file in the configuration file or the commandline parameters.\n");
            return SCAS_ERR_CONFIG_FAIL;
        }
        else
            scas_log(LOG_DEBUG, "Using Public certificate: %s\n", xacml_io_certFilePub); 
    } 

    /* Host / service private key file */
    if (!xacml_io_certFilePriv)
    {
        if (!(xacml_io_certFilePriv = get_scas_config_value_by_name (cnv, n_config_name_values, "scas_hostkey")))
        {
            fprintf (stderr, "SCAS: Couldn't find the setting for the Host / Service private key file in the configuration file or the commandline parameters.\n");
            scas_log (LOG_ERR, "SCAS: Couldn't find the setting for the Host / Service private key file in the configuration file or the commandline parameters.\n");
            return SCAS_ERR_CONFIG_FAIL;
        }
        else
            scas_log(LOG_DEBUG, "Using Private key: %s\n", xacml_io_certFilePriv); 
    } 


    /* Test portnummer appliance, only use first 1024 ports if your root */
    if ((geteuid() != 0) && (portnumber < 1025))
    {
        fprintf (stderr, "SCAS: Without being Root unable to use a port below 1024. Requested portnumber to use: %d\n", portnumber);
        scas_log (LOG_ERR, "SCAS: Without being Root unable to use a port below 1024. Requested portnumber to use: %d\n", portnumber);
        return SCAS_ERR_UNPRIV_PORT;
    }


    /* Properly daemonize the service (if not disabled) */
    /* Any return value other then zero is not good */
    /* Check point */
    if (disableDaemonize)
    {
        fprintf (stderr, "SCAS: Disable daemonization\n");
        scas_log (LOG_INFO, "SCAS: Disable daemonization\n");
	
	/* Setup signal handlers */
	setup_signal_handlers();

	/* Write the pidfile */
	if (write_pidfile())
	    return SCAS_ERR_PIDFILE;
    }
    else
    {
	pid_t pid=daemonize();
        /* if (!daemonize() && !checkpoint()) */
        if (pid<0)
        {
            fprintf (stderr, "SCAS: Couldn't perform the daemonization.\n");
            scas_log (LOG_ERR, "SCAS: Couldn't perform the daemonization.\n");

            return SCAS_ERR_SYSCALLS;
        }
	if (pid>0)  { /* Free memory for parent */
	    /* Cleanup memory from config file */
	    free_config_name_value (&cnv, n_config_name_values);
	    /* Close logfile */
	    scas_log_close();

	    /* exit */
	    _exit(0);
	}
	/* Log a notice that we have started */
	scas_log(LOG_NOTICE,
		"Successfully daemonized, SCAS master process is %ld\n",
		(long)getpid());

	/* Setup signal handlers */
	setup_signal_handlers();

	/* Write the pidfile */
	if (write_pidfile())
	    return SCAS_ERR_PIDFILE;

        /* Start checkpointing from here, and revive the new children here */   
        if (checkpoint())
        {
            scas_log(LOG_ERR, "SCAS: Couldn't perform the checkpointing.\n");

	    /* Cleanup memory from config file */
	    free_config_name_value (&cnv, n_config_name_values);
	    /* Close logfile */
	    scas_log_close();

            return SCAS_ERR_SYSCALLS;
        }
    }

    /* Only start processing XACML in child process or non-daemonize, both cases
     * mychild==0 */
    if (mychild==0) {
	/* Start an application wide mutex, used to channel the LCMAPS execution
	 * because its not thread safe yet */
	pthread_mutex_init(&mutex, NULL);

	scas_log(LOG_DEBUG, "Starting process_xacml\n");
	rc = process_xacml (portnumber);

	pthread_mutex_destroy (&mutex);
    }
    
    /* Cleanup memory from config file */
    free_config_name_value (&cnv, n_config_name_values); 

    /* Cleanup OpenSSL memory */
    xacml_io_cleanup_OpenSSL();

    if (mychild!=0 || disableDaemonize)	{
	if (rc)	{
	    fprintf(stderr,
		    "SCAS finished with non-zero exit value %d\n",rc);
	    scas_log(LOG_NOTICE,
		    "SCAS finished with non-zero exit value %d\n",rc);
	} else {
	    fprintf(stderr,
		    "SCAS finished successfully\n");
	    scas_log(LOG_NOTICE,
		    "SCAS finished successfully\n");
	}
    }

    scas_log_close();
    return rc;
}



/********************************************************************
function:       process_xacml
description:
                This function is the start of the service. It initializes the XACML library
                and will put the library in listen mode. In this case that I/O layer is hardwired to be secured
                via my I/O handler function implementation, that brings you the SSL layer. And in effect, HTTPS.

paramters:
      int listenport    :    The service will listen on this TCP port number
      int secured       :    This is the switch between secured (via I/O handlers) or plain text HTTP. 
********************************************************************/ 
/*int process_xacml (int listenport, int secured)*/
static int process_xacml (int listenport)
{
    int rc = 0;
    const char * logstr = "process_xacml";
    xacml_server_t server;
    /* xacml_response_t response; */
#ifdef HAVE_XACML_SERVER_SET_FD
    int master_socket = -1;
#else
    unsigned short port;

    if (listenport<0 || listenport>USHRT_MAX)	{
	scas_log(LOG_ERR, "%s: Error: Cannot convert port %d to unsigned short\n",
		logstr, listenport);
        return 1;
    }
    port=(unsigned short)listenport;
#endif /* HAVE_XACML_SERVER_SET_FD */



    /******* SSL Init *********/
    if (xacml_io_init_OpenSSL())
    {
        scas_log(LOG_ERR, "%s: Error: Failed to initialize the OpenSSL library\n", logstr);
        return 1;
    }
    else
    {
        scas_log(LOG_DEBUG, "%s: OpenSSL library initialized\n", logstr);
    }

    xacml_io_seed_prng();


    xacml_io_ssl_ctx = xacml_io_setup_server_ctx(NULL, xacml_io_certCAPath, xacml_io_certFilePub, xacml_io_certFilePriv);
    if (!xacml_io_ssl_ctx)
    {
        scas_log(LOG_ERR,     "%s: Error: unable to create an SSL_CTX context with xacml_io_setup_server_ctx()\n", logstr);
        fprintf (stderr, "%s: Error: unable to create an SSL_CTX context with xacml_io_setup_server_ctx()\n", logstr);

	/* not ok */
        rc = 1;
        goto process_openssl_cleanup;
    }
    /**************************/


    xacml_init();
    xacml_server_init(&server, xacml_authorize, NULL);

    /* Set server object properties to listen to give en port */


#ifdef HAVE_XACML_SERVER_SET_FD
    /* Setting up master socket */
    if (xacml_io_setup_master_socket (&master_socket, listenport))
    {
        scas_log(LOG_ERR,     "SCAS setup failure: unable to setup listening socket on port %d\n", listenport);
        fprintf (stderr, "SCAS setup failure: unable to setup listening socket on port %d\n", listenport);

        /* not ok */
        rc = 1;
        goto process_xacml_cleanup;
    }


    /* Setting the listen()-ing master socket. */
    if (xacml_server_set_fd (server, master_socket) != XACML_RESULT_SUCCESS)
    {
        scas_log(LOG_ERR,     "SCAS setup failure: Unable to register the listening socket into the SAML2-XACML2 library.\n");
        fprintf (stderr, "SCAS setup failure: Unable to register the listening socket into the SAML2-XACML2 library.\n");

        /* not ok */
        rc = 1;
        goto process_xacml_cleanup;
    }
        
#else
    xacml_server_set_port(server, (unsigned short)listenport);

    xacml_server_get_port(server, &port);
#endif /* HAVE_XACML_SERVER_SET_FD */

    /* Determine and store the issuer for the XACML responses, now that we have
     * a listening port */
    set_xacml_response_issuer(listenport);

    /* Set the I/O handler object */
    xacml_server_set_io_descriptor(server, &xacml_io_ssl_descriptor);
     
    rc=xacml_server_start(server);

    switch(rc)  {
	case XACML_RESULT_SUCCESS:
#ifndef HAVE_XACML_SERVER_SET_FD
	    scas_log (LOG_NOTICE, "SCAS ready... listening on port %hu.\n", port);
	    printf   ("SCAS ready... listening on port %hu.\n", port);
#endif
	    while (!done)
		pause();

	    /* Perfect ending: make sure rc is really 0 */
	    rc = 0;

	    break;
	case XACML_RESULT_INVALID_PARAMETER:
	    scas_log(LOG_ERR, "xacml_server_start() failed: "
		    "Invalid parameter.\n");
	    break;
	case XACML_RESULT_OBLIGATION_FAILED:
	    scas_log(LOG_ERR, "xacml_server_start() failed: "
		    "Obligation failed.\n");
	    break;
	case XACML_RESULT_SOAP_ERROR:
	    scas_log(LOG_ERR, "xacml_server_start() failed: "
		    "Soap error.\n");
	    break;
	case XACML_RESULT_INVALID_STATE:
	    if (errno)
		scas_log(LOG_ERR, "xacml_server_start() failed: %s\n",
			strerror(errno));
	    else
		scas_log(LOG_ERR, "xacml_server_start() failed:  "
			"Invalid state.\n");
	    break;
	default:
	    scas_log(LOG_ERR, "xacml_server_start() ended with "
		    "unknown error.\n");
	    break;
    }

    if (gsig>0)	{
	fprintf(stderr,
		"SCAS: process_xacml: Shutting down... signal was: %d\n", gsig);
	scas_log(LOG_INFO,
		"process_xacml: Shutting down... signal was: %d\n", gsig);
    } else {
	fprintf(stderr,
		"SCAS: process_xacml: Shutting down...\n");
	scas_log(LOG_INFO,
		"process_xacml: Shutting down...\n");
    }

#ifdef HAVE_XACML_SERVER_SET_FD
process_xacml_cleanup:
#endif /* HAVE_XACML_SERVER_SET_FD */
    /* Destroy server here */
    xacml_server_destroy(server);

    if (errno)
	scas_log(LOG_WARNING,"Server ended in failure: %s\n",
		strerror(errno));

process_openssl_cleanup:
    SSL_CTX_free (xacml_io_ssl_ctx);

    fflush(stdout);
    fflush(stderr);

    return rc;
}

/*
 * Opens (truncates/creates) pidfile, writes current pid in it and closes it
 * again. It uses the location SCAS_PIDFILE.
 * Return -1 on error, 0 on success.
 */
static int write_pidfile(void)	{
    const char *pidfile=SCAS_PIDFILE;
    FILE *pidfilep;
    int rc=0;

    /* Open file */
    if ( (pidfilep=fopen(pidfile,"w"))==NULL)	{
	scas_log(LOG_ERR,"Cannot open pidfile %s: %s\n",
		pidfile,strerror(errno));
	return -1;
    }

    /* Write file */
    if (fprintf(pidfilep,"%ld",(long)getpid())<0)	{
	scas_log(LOG_ERR,"Writing pidfile %s failed: %s\n",
		pidfile,strerror(errno));
	rc=-1;
    }
    
    /* Close file */
    if (fclose(pidfilep)!=0)	{
	scas_log(LOG_ERR,"Writing pidfile %s failed: %s\n",
		pidfile,strerror(errno));
	return -1;
    }

    return rc;
}
